// Prototypes for linear algebra routines from BLAS and LAPACK
// Prototypes for wrapper routines also provided
// All matrices are for cdouble precision complex matrices

#include <iostream>
#include <vector>
#include "Type.h"
#include <complex.h>
#include <cmath>    // round etc


using namespace std;

// Extern "C" statement needed to prevent C++ decoration of function names
extern "C" {
	// BLAS routine to copy all/part of a matrix
	void zlacpy_(const char *UPLO, int *M, int *N, const cdouble *A, int *LDA, cdouble *B, int *LDB);

	// LU decomposition
	void zgetrf_(int *M, int *N, cdouble *A, int *LDA, int *IPIV, int *INFO);

	// Matrix inversion following LU decomposition (zgetrf)
	void zgetri_(int *N, cdouble *A, int *LDA, int *IPIV, cdouble *WORK, int *LWORK, int *INFO);

        // Matrix-vector multiplication (BLAS level 2)
        void zgemv_(const char *TRANS, int *M, int *N, cdouble *ALPHA, const cdouble *A, int *LDA
                , const cdouble *X, int *INCX, cdouble *BETA, cdouble *Y, int *INCY);

        // Matrix-matrix multiplication (BLAS Level 3)
        void zgemm_(const char *TRANSA, const char *TRANSB, int *M, int *N, int *K, double *ALPHA
                , const cdouble *A, int *LDA, const cdouble *B, int *LDB, double *BETA, cdouble *C, int *LDC);

        // Determine eigenvalues and eigenvectors
        void zgeev_(const char *JOBVL, const char *JOBVR, int *N, cdouble *A, int *LDA, cdouble *W
                , cdouble *VL , int *LDVL, cdouble *VR, int *LDVR, cdouble *WORK, int *LWORK
                , double *RWORK, int *INFO);

        // Determine eigenvalues and eigenvectors
        void dgeev_(const char *JOBVL, const char *JOBVR, int *N, double *A, int *LDA, double *WR, double *WI
                , double *VL , int *LDVL, double *VR, int *LDVR, double *WORK, int *LWORK
                , int *INFO);


}

int idxr(int c, int r, int order);
int idxi(int c, int r, int order);
int idvr(int c);
int idvi(int c);

// B = A
void zmat_cpy(int order, const cdouble *A, cdouble *B);

// A-1
int zmat_inv(int order, const cdouble *mat, cdouble *imat);

// C = A x B
void zmat_mlt(int order, const cdouble *A, const cdouble *B, cdouble *C);

// D = A x B x C
void zmat_mlt(int order, const cdouble *A, const cdouble *B, const cdouble *C, cdouble *D);

// Compute eigenvalues
int zmat_eig(int order, cdouble *mat, cdouble *evec, cdouble *eval);

// Compute right/left eigenvectors and eigenvalues
int dmat_eig(int order, double *mat, double *eigReal, double *eigImag);

// Compute eigenvalues and left and right eigenvectors
int zmat_evv(int order, cdouble *mat, cdouble *vr, cdouble *vl , cdouble *eval);

// A**H complex conjugate
void zmat_cnj(int order, const cdouble *A, cdouble *cnjA);

// matx C = matx A + matx B 
void zmat_add( int order, const cdouble *A, cdouble *B, cdouble *C );

// matx D = matx A + matx B + matx C 
void zmat_add( int order, const cdouble *A, cdouble *B, cdouble *C , cdouble *D );

// round matrix elements 
void zmat_rnd(int order, cdouble *A, cdouble *B);

// B = exp( -1 * t * eigVal ) // diagonal matrix only!
void zmat_exp(int order, double t, const cdouble *A, cdouble *B);

// Y = A * X ( matrix A and vector X )
void zmat_vmlt(int order, const cdouble *A, const cdouble *X, cdouble *Y);

// C = A * B ( row_vector A and  matrix B )
void zmat_rvm(int order, const cdouble *A, const cdouble *B, cdouble *C);

// Y = A * X  (row_vector A and column vector X)
void zmat_rvv(int order, const cdouble *A, const cdouble *X, cdouble *Y);


