
  #include <iostream>
  #include <iomanip>
  #include <fstream>
  #include <complex.h>
  #include "Type.h"
  #include <dlib/optimization.h>
  #include "Read.h"
  #include "Struc.h"
  #include "Stats.h"

// ------------------------------------------------------------------
// 2 spin density matrix ; removed exchange 

 using namespace std;

 int main(){

  if( fileExists( "2d.txt" ) ) { ; } else { print_help() ; }            // check if file exits else print help
  if( fileExists( "peaks.txt" ) ) { ; } else { print_help() ; }

  params_type k = { 0.2 , 0.2 , 1. , 3.  } ;				// fitting params in dlib vector type
  params P ;                                     			// define parameter structure

  const char* PEAKS = "peaks.txt";                                  	// define file to data file to read-in
  cout << endl << "# Reading \"peaks.txt\" file. " << endl ;
  std::vector< std::vector<double> > A = fileIn( PEAKS );           	// push data file in matrix A / see Read.h 

  P.deltaA = A[0][0] ;							// assign peaks to fit
  P.deltaB = A[0][1] ;

  double high , low ;							// sort peak range low to high
  if ( A[1][0] > A[1][1] ) { high = A[1][0] ; low = A[1][1] ; }
  else { low = A[1][0] ; high = A[1][1] ; }

  const char* FILE_NAME = "2d.txt";                                 	// define data file to read-in
  vector<vector<vector<double>>> DATA = fileIn3( FILE_NAME );      	// push data file into matrix A / see Read.h 
  cout << "# Reading \"2d.txt\" file. " << endl ;

// ------------------------------------------------------------------
// Messy routine to cut the data according peak range in peaks.txt  

// find max peak first
  double max = 0. ;
  for ( int j = 0 ; j < DATA.size() ; j++ ) {
    for ( int k = 0 ; k < DATA[j].size() ; k++ ) {
      if ( DATA[j][k][0] > low && DATA[j][k][0] < high && DATA[j][k][1] > max ) {
        max = DATA[j][k][1]  ;
      }
    }
  }

  vector<vector<vector<double>>> CUT3 ;      				// push data file into matrix A / see Read.h 
  vector<vector<double>> CUT2 ;      					// push data file into matrix A / see Read.h 
  for ( int j = 0 ; j < DATA.size() ; j++ ) {
    for ( int k = 0 ; k < DATA[j].size() ; k++ ) {
      if ( DATA[j][k][0] > low && DATA[j][k][0] < high ) {
        CUT2.push_back( { DATA[j][k][0] , DATA[j][k][1] / max } ) ;
      }
    }
        CUT3.push_back( CUT2 ) ;
	CUT2.clear() ;
  }

  vector<vector<vector<double>>> RES3 ;      	// define simulated NMR spectrum / ppm 
  vector<vector<vector<double>>> SIM3 ;      	// define simulated NMR spectrum / ppm 
  vector<vector<vector<double>>> SIM3A ;      	// define simulated NMR spectrum / ppm 
  vector<vector<vector<double>>> SIM3B ;      	// define simulated NMR spectrum / ppm 
  vector<vector<double>> FIN2 ;      		// define simulated NMR spectrum / ppm 
  vector<double> FIN ;      			// define simulated NMR spectrum / ppm 

// ------------------------------------------------------------------

  ofstream file2a;
  ofstream file2b;
//  file2.open("sim.txt");
  cout << "# Writing simulation \"sim.txt\" file. " << endl ;

  ofstream file3;
//  file3.open("res.txt");
  cout << "# Writing residuals \"res.txt\" file. "  << endl ;

  ofstream file4;
  //file4.open("results.txt");
  cout << "# Writing results \"results.txt\" file. " << endl ;

  ofstream file5;
//  file5.open("dat.txt");
  cout << "# Writing scaled original data \"dat.txt\" file. " << endl << endl ;

  cout << "# index MA MB LWA LWB " << endl ;

  for ( int j = 0 ; j < DATA.size() ; j++ ) {

// ------------------------------------------------------------------
// lambda expression below! &DATA is "captured"
// note it ends with }; and not }

  auto density_matrix = [ &P , &DATA , &j , &high , &low , &CUT3 , &SIM3 , &SIM3A , &SIM3B , &RES3 , &FIN2 , &FIN ]( const params_type& params_sim ) {

  std::vector< std::vector<double> > SIM ;      // define simulated NMR spectrum / ppm 

  P.MA  = params_sim(0,0) ;
  P.MB  = params_sim(0,1) ;
  P.R2A = params_sim(0,2) ;
  P.R2B = params_sim(0,3) ;

// ------------------------------------------------------------------
// Calc FID

  cdouble A ;
  cdouble numA , denA ;
  cdouble numB , denB ;

  for ( int k = 0 ; k < CUT3[j].size() ; k++ ) {
    numA = P.MA * ( 1. / P.R2A ) ;
    numB = P.MB * ( 1. / P.R2B ) ;
    denA = 1 + pow ( ( CUT3[j][k][0] - P.deltaA ) , 2 ) * pow ( ( 1. / P.R2A) , 2 ) ;
    denB = 1 + pow ( ( CUT3[j][k][0] - P.deltaB ) , 2 ) * pow ( ( 1. / P.R2B) , 2 ) ;

    A = ( numA / denA ) + ( numB / denB ) ;

    SIM.push_back( { real(CUT3[j][k][0]) , real(A) } ) ;

  }

  return calc_residuals ( SIM , CUT3[j] ) ; 

 } ;  // end lambda function

// ------------------------------------------------------------------
// dlib minimize residuals routine

  dlib::find_min_box_constrained(dlib::bfgs_search_strategy(),
                             dlib::objective_delta_stop_strategy(1e-15),
                             density_matrix, dlib::derivative(density_matrix), k, 0.0001, 10. );

// print results to screen
  cout  << fixed << setw(2) << setprecision(0) 
        << j         
        << fixed << setw(2) << setprecision(3) 
	<< " " << k(0,0) << " " << k(0,1)   
        << " " << k(0,2) << " " << k(0,3) << endl ; 

  for ( int a = 0 ; a < k.size() ; a++ ) {
    FIN.push_back(  k(0,a) ) ;
  }

// ------------------------------------------------------------------

  cdouble A , B ;
  cdouble numA , denA ;
  cdouble numB , denB ;
  cdouble intA = {0,0} ;
  cdouble intB = {0,0} ;

  vector<vector<double>> SIM2A ;      // define simulated NMR spectrum / ppm 
  vector<vector<double>> SIM2B ;      // define simulated NMR spectrum / ppm 
  vector<vector<double>> SIM2 ;      // define simulated NMR spectrum / ppm 
  vector<vector<double>> RES2  ;      // define simulated NMR spectrum / ppm 
  for ( int k = 0 ; k < CUT3[j].size() ; k++ ) {

    numA = P.MA * ( 1. / P.R2A ) ;
    numB = P.MB * ( 1. / P.R2B ) ;
    denA = 1 + pow ( ( CUT3[j][k][0] - P.deltaA ) , 2 ) * pow ( ( 1. / P.R2A) , 2 ) ;
    denB = 1 + pow ( ( CUT3[j][k][0] - P.deltaB ) , 2 ) * pow ( ( 1. / P.R2B) , 2 ) ;

    A = ( numA / denA ) ;
    B = ( numB / denB ) ;
    intA = intA + A ;			// integrate A and B
    intB = intB + B ;

  SIM2A.push_back( { CUT3[j][k][0] , real(A) } ) ;	// store simultated A and B
  SIM2B.push_back( { CUT3[j][k][0] , real(B) } ) ;
  SIM2.push_back( { CUT3[j][k][0] , real(A) + real(B) } ) ;
  RES2.push_back( { CUT3[j][k][0] , ( CUT3[j][k][1] - real(A) - real(B) ) } ) ;	// store residuals
  }
  SIM3A.push_back ( SIM2A ) ;
  SIM3B.push_back ( SIM2B ) ;
  SIM3.push_back ( SIM2 ) ;
  RES3.push_back ( RES2 ) ;
  SIM2A.clear() ;
  SIM2B.clear() ;
  RES2.clear() ;
  FIN.push_back( real(intA) ) ;
  FIN.push_back( real(intB) ) ;
  FIN.push_back( calc_residuals ( SIM3[j] , CUT3[j] ) ) ;
  FIN2.push_back( FIN ) ;
  FIN.clear() ;

 }

   cout << endl << endl ;

// ------------------------------------------------------------------
// print most everything to files named below

  file5.open ("dat.txt");
  file2a.open("sim_a.txt");
  file2b.open("sim_b.txt");
  file3.open ("res.txt");
  file4.open ("results.txt");

  //print_results (FIN2) ;
  file4 << "# index height_A height_B hwhh_A hwhh_B integ_A integ_B residual_square_tot " << endl ;
  for( long unsigned int i = 0 ; i < FIN2.size() ; i++ )    {
    file4 << fixed << setw(2) << setprecision(0) <<  i << " " ;
  for( long unsigned int j = 0 ; j < FIN2[i].size() ; j++ ) {
    file4 << fixed << setw(6) << setprecision(3) <<  FIN2[i][j] << " " ;
        }
                file4 << endl ;
    }

  for ( int k = 0 ; k < CUT3[0].size() ; k++ ) {
    for ( int j = 0 ; j < CUT3.size() ; j++ ) {
      if ( j < 1 ) { 
	file5 << fixed << setw(10) << setprecision(6) << CUT3[j][k][0] << " " ; 
	file2a << fixed << setw(10) << setprecision(6) << CUT3[j][k][0] << " " ; 
	file2b << fixed << setw(10) << setprecision(6) << CUT3[j][k][0] << " " ; 
	file3 << fixed << setw(10) << setprecision(6) << CUT3[j][k][0] << " " ; 
	}
      file5  << fixed << setw(10) << setprecision(6)
             << CUT3[j][k][1] << " "  ;
      file2a << fixed << setw(10) << setprecision(6)
             << SIM3A[j][k][1] << " "  ;
      file2b << fixed << setw(10) << setprecision(6)
             << SIM3B[j][k][1] << " "  ;
      file3  << fixed << setw(10) << setprecision(6)
             << RES3[j][k][1] << " "  ;
    }
    file5  << endl ; 
    file2a << endl ; 
    file2b << endl ; 
    file3  << endl ; 
  }

  file5.close();
  file2a.close();
  file2b.close();
  file3.close();
  file4.close();

  return 0;

}

// ------------------------------------------------------------------
// ------------------------------------------------------------------
// ------------------------------------------------------------------
