
#include <vector>
#include <iomanip>
#include <fstream>
#include <vector>
#include <cmath>
#include <iostream>
#include <sstream>
#include <dlib/optimization.h>

using namespace std;

// ----------------------------------------------------------------------------------------
// read in matrix / nmr data of variable size
   std::vector< std::vector<double> > fileIn( const char* filename )
{
   std::vector< std::vector<double> > arr;
   ifstream in( filename );
   if ( in )
   {
      for ( string line; getline( in, line ); )
      {
      if( (line[0] == '#') | (line[0] == '&') | (line.empty()) | (line[0] == '\n') ) continue; // skip #, & and empty lines 
         stringstream ss( line );
         std::vector<double> row;
         for ( double i; ss >> i; ) row.push_back( i );
         arr.push_back( row );
      }
   }
   else
   {
      cerr << "No such read-in file\n";
   }
   return arr;
}

// ----------------------------------------------------------------------------------------
// read in matrix / nmr data of variable size pseudo 2D mode

  std::vector< std::vector< std::vector<double> > > fileIn3( const char* filename )
{
   std::vector< std::vector< std::vector<double> > > cnt;
   std::vector< std::vector<double> > arr;
   ifstream in( filename );
   if ( in )
   {
      for ( string line; getline( in, line ); )
      {
      if( (line[0] == '#') | (line.empty()) | (line[0] == '\n') ) continue; // skip #, & and empty lines 
      if( (line[0] == '&') ) { cnt.push_back( arr ); arr.clear() ; continue ; }  // store 3d vector and clear 2d vec  
         stringstream ss( line );
         std::vector<double> row;
         for ( double i; ss >> i; ) row.push_back( i );
         arr.push_back( row );
      }
   }
   else
   {
      cerr << "No such read-in file\n";
   }
   return cnt;
}


// ----------------------------------------------------------------------------------------
struct Point
{
    double x;
    double y;
};

// ----------------------------------------------------------------------------------------
// Any distance function can be used for optimisation.  This one, where we want
// to find the least-squares is most common. 

double dist(double y1, double y2)
{
    double y = y2 - y1;
    return (y * y);
}

// ----------------------------------------------------------------------------------------
   
  void print_time ( const vector<vector<double>> &A)
  {
  for( long unsigned int i = 0 ; i < A.size() ; i++ )    {
    cout << fixed << setw(8) << setprecision(6) <<  A[i][0] << " " ;
    cout << endl ;
        }
  }

// ----------------------------------------------------------------------------------------
// print initial data 

  void print_DATA ( const vector<vector<double>> &A)
  {
  for( long unsigned int i = 0 ; i < A.size() ; i++ )    {
    cout << fixed << setw(2) << setprecision(0) <<  i << " " ;
  for( long unsigned int j = 0 ; j < A[i].size() ; j++ ) {
    cout << fixed << setw(6) << setprecision(3) <<  A[i][j] << " " ;
        }
                cout << endl ;
    }
  }


// ----------------------------------------------------------------------------------------
// print initial data pseudo 2d mode 
  void print_DATA3 ( const vector<vector<vector<double>>> &A)
  {
  for( long unsigned int a = 0 ; a < A.size() ; a++ ) {
  for( long unsigned int b = 0 ; b < A[a].size() ; b++ ) {
  for( long unsigned int c = 0 ; c < A[a][b].size() ; c++ ) {

    cout << fixed << setw(8) << setprecision(6)
         << A[a][b][c] << " " ;
   }
                cout << endl ;
  }
                cout << "&" << endl;
 }
}

// ----------------------------------------------------------------------------------------
// convert to dlib matrix
// set number of rate constants below too

  dlib::matrix<double, 2 , 1 > vec2matrix ( const vector<vector<double>> &V )
  {
  dlib::matrix<double, 2 , 1 > C ;
  for( long unsigned int i = 0 ; i < V.size() ; i++ )    {
  for( long unsigned int j = 0 ; j < V[i].size() ; j++ ) {
    C(i,j) = V[i][j] ;
        }
    }
   return C ;
  }

// Function to check if a file exists
bool fileExists( const std::string& filename ) {
    //if (std::ifstream fileStream(filename.c_str()) )
     std::ifstream fileStream(filename.c_str()); 
    return fileStream.good();
}

// Function to check if a file exists
  void print_help ( )
  {
		cerr <<
                 "##############################################\n"
                 "##############################################\n\n"
                 "ERROR: Not enough input files specified. \n"
                 "USAGE: reads in 'peak.txt' and '2d.txt'  \n"
                 " files by default so they must be in your directory  \n"
                 "\n"
                 " peak.txt format: \n"
                 " peak_A(in ppm) peak_B(in ppm) \n"
                 " peak range to cut in ppm \n"
                 "  \n"
                 " eg): \n"
                 " 0.013 0.234 \n"
                 " -5.2 5.2 \n\n"
                 "##############################################\n"
                 "##############################################\n"
		;
	exit (EXIT_FAILURE) ;
 }

// ----------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------
