// Type.h		Various typedef declarations
// Kenji Sugase				10/9/2012

#ifndef __TYPES_h__
#define __TYPES_h__

#include <vector>
#include <string>
#include <complex> 
#include <dlib/optimization.h>


using namespace std;

typedef vector<int>			ivect;
typedef vector<vector<int> >		i2vect;
typedef vector<vector<vector<int> > >	i3vect;
typedef vector<double>			dvect;
typedef vector<vector<double> >		d2vect;
typedef vector<string>			svect;
typedef vector<vector<string> >		s2vect;
typedef vector<bool>			bvect;
typedef vector<vector<bool> >		b2vect;
typedef complex<int>			cint;
typedef complex<double>			cdouble;

typedef dlib::matrix< double , 4 , 1 > params_type ;           // define params type vector with the number of rate constants


// Trick to allow type promotion below
// allows multiply imaginary number i * an integer
// eg) i*5     without error.  otherwise need i*5.  to call a double
template <typename T>
struct identity_t { typedef T type; };

/// Make working with std::complex<> nubmers suck less... allow promotion.
#define COMPLEX_OPS(OP)                                                 \
  template <typename _Tp>                                               \
  std::complex<_Tp>                                                     \
  operator OP(std::complex<_Tp> lhs, const typename identity_t<_Tp>::type & rhs) \
  {                                                                     \
    return lhs OP rhs;                                                  \
  }                                                                     \
  template <typename _Tp>                                               \
  std::complex<_Tp>                                                     \
  operator OP(const typename identity_t<_Tp>::type & lhs, const std::complex<_Tp> & rhs) \
  {                                                                     \
    return lhs OP rhs;                                                  \
  }
COMPLEX_OPS(+)
COMPLEX_OPS(-)
COMPLEX_OPS(*)
COMPLEX_OPS(/)
#undef COMPLEX_OPS

struct params
{
        double start;           // first point
        double end;             // last point
        double lb;              // Line Broadening rads
        double sw ;             // Sweep width Hz 
        double swp;             // Sweep width ppm
        double dw ;             // Dwell time     
        double w0 ;             // Dwell time     
        double wrt ;            // vr * t     
        double wr  ;            // vr         
        double dip ;            // vr * t     
        double time;            // vr * t     
        double transmit ;       // Dwell time     
        double ph ;             // phase in rads
        double b1 ;             // applied b1 field
        double tf ;             // pulse length
        double om1 ;            // frequency?     
        double R1  ;            // R1 relaxation  
        double R2  ;            // R2 relaxation  
        double R2A  ;            // R2 relaxation  
        double R2B  ;            // R2 relaxation  
        double k   ;            // Rate constant events/sec
        double kf   ;           // Rate constant events/sec
        double kr   ;           // Rate constant events/sec
        cdouble delta ;         // delta (rads/sec) 
        cdouble deltaA ;       	// delta (rads/sec) 
        cdouble deltaB ;       	// delta (rads/sec) 
        cdouble MA  ;            // Pop A   
        cdouble MB  ;            // Pop B   

};

#endif
