
  #include <iostream>
  #include <iomanip>
  #include <fstream>
  #include <complex.h>
  #include "LinearAlgebra.h"
  #include "Type.h"
  #include <dlib/optimization.h>
  #include "Read.h"
  #include "Struc.h"
  #include "Stats.h"

// ------------------------------------------------------------------
// 2 spin density matrix ; removed exchange 

 using namespace std;

 int main(){

// ------------------------------------------------------------------

  //params_type k = { 1. , 3. , 1. , 3. } ;
  params_type k = { 1. , 3.  } ;					// fitting params in dlib vector type

  const char* FILE_NAME = "dat.txt";                                 	// define data file to read-in
  std::vector< std::vector<double> > DATA = fileIn( FILE_NAME );      	// push data file into matrix A / see Read.h 
  cout << endl << " Reading \"dat.txt\" file. " << endl << endl ;

// ------------------------------------------------------------------
// allocate

  int n = 2 , m = 2 ;

  cdouble *A;           A = new cdouble[n*m*1];
  cdouble *B;           B = new cdouble[n*m*1];
  cdouble *C;           C = new cdouble[n*m*1];
  cdouble *D;           D = new cdouble[n*m*1];
  cdouble *E;           E = new cdouble[n*m*1];
  cdouble *F;           F = new cdouble[n*m*1];
  cdouble *IxS;         IxS = new cdouble[n*m*1];
  cdouble *Sm;          Sm = new cdouble[n*m*1];
  cdouble *S;           S = new cdouble[n*m*1];
  cdouble *eigVal;      eigVal = new cdouble[n*1];

  cdouble *vecf;        vecf = new cdouble[n*2];
  cdouble *M0;          M0 = new cdouble[n*2];
  cdouble *data;        data = new cdouble[n*m*2];

// ------------------------------------------------------------------

  params P ;                                     // define parameter structure

//  P.MA = { 1. , 0. } ;                          // Pop A ; complex number
//  P.MB = { 9. , 0. } ;                         // Pop B ; complex number
    
  //P.deltaA  = params_sim(0,0) ;
  //P.deltaB = params_sim(0,1) ;
  P.deltaA = 1.   ;                             // chem shift of A in PPM
  P.deltaB = -2.14 ;                            // chem shift of B in PPM

  //P.R2A  = params_sim(0,2) * M_PI * 2. ;
  //P.R2B = params_sim(0,3) * M_PI * 2. ;
  P.R2A = 1.0 * M_PI * 2. ;                       // Relaxation times and offset rads/sec?
  P.R2B = 11. * M_PI * 2. ;                       // Relaxation times and offset rads/sec?

  P.transmit = 0.   ;                           // pulse frequency center Hz 
  P.swp = 10.  ;                                // sweepwidth in ppm
  P.w0 = 100.   ;                               // 1H larmor freq
  P.dw = 1./(P.swp * P.w0)  ;                   // dwell time in seconds
  P.sw = P.swp * P.w0 ;

  P.start = 0 ;                                 // 1st fid point
  P.end = 4096 ;                                // number of fid pts

  int end = P.end ;
  cdouble FID[ end * 2 ] ;                      // initialize FID memory space

// ------------------------------------------------------------------
// lambda expression below! &DATA is "captured"
// note it ends with }; and not }

  auto density_matrix_1 = [&P,&FID,&DATA,&A,&B,&C,&D,&E,&F,&IxS,&Sm,&S,&eigVal,&vecf,&M0,&data]( const params_type& params_sim ) {

  std::vector< std::vector<double> > SIM ;      // define simulated NMR spectrum / ppm 

  P.MA  = params_sim(0,0) ;
  P.MB = params_sim(0,1) ;
//  P.R2A = params_sim(0,2) * 2. * M_PI ;
//  P.R2B = params_sim(0,3) * 2. * M_PI ;

// ------------------------------------------------------------------

  const complex<double> im(0.0, 1.0);
  double t ;
  int order = 2 ;

// ------------------------------------------------------------------
// NMR parameters that need to be passed

  cdouble deltaA , deltaB ;

  // chem shifts in rads / sec ; sign flip for some reason

  deltaA = -1. * P.deltaA * P.w0 * im * 2. * M_PI ;                  	// delta in rads/sec
  deltaB = -1. * P.deltaB * P.w0 * im * 2. * M_PI ;                  	// delta in rads/sec

// ------------------------------------------------------------------
// Chemical shift matx

  cdouble omega[2*2*2] = {
        { deltaA }  , { 0. , 0. } ,
        { 0. , 0. } , { deltaB }  
  };

// Relaxation matx
  cdouble relax[2*2*2] = {
        { P.R2A } , { 0. , 0. } , 
        { 0. , 0. } , { P.R2B } 
  };

// Population Vector; 

  M0[0] = P.MA ; 
  M0[1] = P.MB ; 

// ------------------------------------------------------------------
// combine chem shift, relax and exch into one matx

  zmat_add( order, omega, relax, data );

// ------------------------------------------------------------------

  cdouble one[2*2] = {
        { 1. , 0. } , { 1. , 0. } , 
  };

// ------------------------------------------------------------------

  zmat_cpy ( order , data , A ) ;
  zmat_cpy ( order , data , B ) ;
  zmat_eig ( order , A , Sm , eigVal ) ;
  zmat_inv ( order , Sm , S ) ;
  zmat_mlt ( order , S , B  , Sm , D ) ;

// ------------------------------------------------------------------
// Calc FID

  cdouble sum = {0,0} ;
  for ( int k = 0 ; k < P.end ; k++ ) {

    t = -1. * k * P.dw ;
    zmat_exp ( order , t , D , E ) ;
    zmat_mlt ( order , Sm , E , S , F ) ;
    zmat_rvm ( order , one , F , IxS ) ;
    zmat_rvv ( order , IxS , M0  , vecf ) ;

    sum = {0,0} ;
    for ( int j = 0 ; j < order ; j++ ) {
      sum = vecf[j]  + sum ;
    }
      FID[k] =  sum ;
  }

  genFFT3( P, FID , SIM );                            // FFT of FID store VEC SPEC
//  print_DATA ( SIM ) ;
//  print_residuals ( SIM , DATA ) ;
//  print_residuals ( SIM , DATA ) ;

  return calc_residuals ( SIM , DATA ) ; 

//  delete [] A;    delete [] B;    delete [] C;
//  delete [] D;    delete [] E;    delete [] F;
//  delete [] vecf; delete [] data; delete [] M0;
//  delete [] IxS;  delete [] Sm;   delete [] S;
  
  } ;  // end lambda function


// ------------------------------------------------------------------
// dlib minimize residuals routine

  dlib::find_min_box_constrained(dlib::bfgs_search_strategy(),
                             dlib::objective_delta_stop_strategy(1e-7),
                             density_matrix_1, dlib::derivative(density_matrix_1), k, 0. , 22. );

  cout  << " k1  = " << k(0,0) << endl
        << " k1r = " << k(0,1) << endl  ; 
 
  return 0;

  }

// ------------------------------------------------------------------
// ------------------------------------------------------------------
// ------------------------------------------------------------------
