// Struc.h  Structures for Eulers and printing	

#include <stdio.h>
#include <fstream>
#include <iostream>
#include <iomanip>

using namespace std;

#ifndef __STRUC_H__
#define __STRUC_H__

struct euler 
{
	double alpha;		// Alpha angle
	double beta;		// Beta angle
	double gamma;		// Gamma angle
	double weight;		// Weight of list
};

void print_euler (const euler* start, const euler* stop)
{
  const euler * current = start;
  while (current != stop) {
    cout << fixed << setw(8) << setprecision(5) << current->alpha << "\t" 
	          << setw(8) << setprecision(5) << current->beta  << "\t" 
                  << setw(8) << setprecision(7) << current->weight << endl;
    ++current;     // increment pointer
  }
}

// count lines in file exclude spaces and #
int genV ( const char *filename ) 
{
  fstream fin(filename);
  string line;
  int i = 0;
  double a, b;

while(getline(fin, line))
  {
  if(line[0] == '#' | line.empty() ) continue; // skip '#' and empty lines in shape.txt file
  stringstream ss(line);
  if (ss >> a >> b) { }
  i++;
  }
  return i;
}

struct spectrum
{
	double freal;		// Real FID   
	double fimag;		// Imag FID   
};

struct poly 
{
	cdouble a;		// Poly coeff
	cdouble b;		// Poly coeff
	cdouble c;		// Poly coeff
	cdouble d;		// Poly coeff 
};


void print_fid (const double* start, const double* stop)
{
  const double * current = start;
  while (current != stop) {
    //cout << fixed << setw(8) << setprecision(5) << (*current) << endl; 
    cout << *current << endl; 
    ++current;     // increment pointer
  }
}

void genFFT ( const params nmr_params, const spectrum * FID )
  {
  int end = nmr_params.end;
  double FI , FR , FRM , sum , sum_real , sum_imag , phase , xaxis ;
  double du = nmr_params.swp / end ;
  double dw = 1. / nmr_params.swp ;

//for(double k = -1. * end/ 2. ; k < (end-1.)/2. ; k++){
for(double k = -1. * end/ 2. ; k < 0.5 * end-1 ; k++){
  FI = FR = FRM = sum = sum_real = sum_imag = 0. ;
for(int m = 0; m < end; m++){
  //phase = M_PI * k * ( m + 0.5 ) / end;	// gives phase twists not sure its right
  phase = 2 * M_PI * k *  m / end;
  FR = FID[m].freal * ( cos ( phase ) ) ;
  FI = FID[m].fimag * ( sin ( phase ) ) ;
  sum_real = FR + sum_real ;
  sum_imag = FI + sum_imag ;
  sum = sum_real + sum_imag ;
  //sum = sum_real  ;
}
  sum = ( sum / end ) ;
  sum_real = sum_real / end ;
  sum_imag = sum_imag / end ;
  xaxis = ( k * du + nmr_params.transmit ) ;

  cout << fixed << setw(20) << setprecision(8) << xaxis
       << fixed << setw(20) << setprecision(8) << sum_real << endl;

  sum = sum_real = sum_imag = phase = FR = FI = 0. ;
  }
}

void genFFT2 ( const params nmr_params, const cdouble * FID )
  {
  double end = nmr_params.end;
  cdouble sum , phase ;
  double xaxis ;
  double du = nmr_params.swp / end ;
  double dw = 1. / nmr_params.swp ;
  const complex<double> im(0.0,1.0);

for( double k = (-0.5 * end) ; k < (0.5 * end -1. ) ; k++){
  sum = {0,0} ;
for( int m = 0 ; m < end ; m++ ) {
  phase = ( 2. * M_PI * k * m ) / end ;
  sum = sum + FID[m] * exp( -1. * im * phase ) ;
}
  sum = ( sum / end ) ;
  xaxis = ( k * du + nmr_params.transmit ) ;

  cout << fixed << setw(20) << setprecision(8) << xaxis
       << fixed << setw(20) << setprecision(8) << real(sum) << endl;
  sum  = phase = {0,0} ;
  }
}

// fft and store spectrum into 2d vector
void genFFT3 ( const params nmr_params, const cdouble * FID , vector< vector<double> > &SPEC )
  {
  double end = nmr_params.end;
  cdouble sum , phase ;
  double xaxis ;
  double du = nmr_params.swp / end ;
  double dw = 1. / nmr_params.swp ;
  const complex<double> im(0.0,1.0);
    
for( double k = (-0.5 * end) ; k < (0.5 * end -1. ) ; k++){
  sum = {0,0} ;
for( int m = 0 ; m < end ; m++ ) {
  phase = ( 2. * M_PI * k * m ) / end ;
  sum = sum + FID[m] * exp( -1. * im * phase ) ;
} 
  sum = ( sum / end ) ;
  xaxis = ( k * du + nmr_params.transmit ) ;

  SPEC.push_back({ xaxis, real(sum) , imag(sum) });
  sum  = phase = {0,0} ;
  }
}

void print_FFT ( vector< vector<double> > &SPEC )
  {
    for( int i = 0 ; i < SPEC.size() ; i++ ) {
    for( int j = 0 ; j < SPEC[i].size() ; j++ )
      cout << fixed << setw(8) << setprecision(6) << SPEC[i][j] << " " ;
      cout << endl;
  }
}

void print_dip ( const params nmr_params, const spectrum * FID )
  {
  int end = nmr_params.end; 

for(int m = 0; m < end; m++){
    cout << fixed << setw(14) << setprecision(9) << FID[m].freal / FID[0].freal<< endl; 
  }
}

void print_fid2 ( const params nmr_params, const cdouble * FID )
  {
  int end = nmr_params.end; 
  double xaxis ;
  double dw=1/(nmr_params.swp);

for(int m = 0; m < end; m++){
  xaxis = m * dw ;
  cout << fixed << setw(10) << setprecision(5) << xaxis << " "      
       << fixed << setw(14) << setprecision(9) << FID[m] << endl; 
  }
}

void print_fid_real ( const params nmr_params, const spectrum * FID )
  {
  int end = nmr_params.end; 
  double xaxis ;
  double dw=1/(nmr_params.swp);

for(int m = 0; m < end; m++){
  xaxis = m * dw ;
  cout << fixed << setw(10) << setprecision(5) << xaxis        
       << fixed << setw(14) << setprecision(9) << FID[m].freal << endl; 
  }
}

void print_fid_imag ( const params nmr_params, const spectrum * FID )
  {
  int end = nmr_params.end; 
  double xaxis ;
  double dw=1/(nmr_params.swp);

for(int m = 0; m < end; m++){
  xaxis = m * dw ;
  cout << fixed << setw(10) << setprecision(5) << xaxis        
       << fixed << setw(14) << setprecision(9) << FID[m].fimag << endl; 
  }
}

void print_cart ( double (*arr)[3][3] )
{
 for (int i = 0; i < 3 ; i++) {
 for (int j = 0; j < 3 ; j++) {
    cout << fixed << setw(8) << setprecision(1) << (*arr)[i][j] << " " ;
    }
  cout << endl;
  }
}

void print_cart_complex ( cdouble (*arr)[3][3] )
{
 for (int i = 0; i < 3 ; i++) {
 for (int j = 0; j < 3 ; j++) {
    cout << fixed << setw(8) << setprecision(1) << (*arr)[i][j] << " " ;
    }
  cout << endl;
  }
}

// count lines in file exclude spaces and #
// store nmr params b1 and ph
int genShape ( params * param, const char *filename ) 
{
  fstream fin(filename);
  string line;
  int i = 0;
  double a, b;

while(getline(fin, line))
  {
  if(line[0] == '#' | line.empty() ) continue; // skip '#' and empty lines in shape.txt file
  stringstream ss(line);
  if (ss >> a >> b) { }
  param[i].b1 = a;
  param[i].ph  = b;
//  cout << param[i].b1 << " " << param[i].ph << endl;
  i++;
  }
  return i;
}

void print_matx(int order, const cdouble *A) {
  int m = order; 
  int n = order; 
  for (int j=0;j<m;j++){
    for (int i=0;i<n;i++){
        cout << fixed << setw(8) << setprecision(2) << A[j*n+i] << " " ;
    }
        cout << endl ;
  }
        cout << endl ;
}

void print_vec(int order, const cdouble *A) {
  int n = order; 
    for (int i=0;i<n;i++){
        cout << fixed << setw(8) << setprecision(4) << A[i] << endl ;
    }
      cout << endl;
}

void norm_vec(int order, cdouble *A) {
  int n = order; 
  cdouble sum = 0. ;
    for (int i=0;i<n;i++){
        sum = sum + A[i]  ;
    }
    for (int i=0;i<n;i++){
        A[i] = A[i] / sum  ;
    }
}

// can determine pops from eig vector who's eig val = 0.  important see bain review
// should only be one eig val = 0 else set manually if 2 or more.
void norm_mag( const int order,  const cdouble *eval, const cdouble *evec, cdouble *M0) {
  int n = order ; 
  int k = order;
  int count = 0 ; 
  double ev_real = 0. ;
  cdouble sum = {0,0} ;

    for ( int i = 0 ; i < order ; i++ ) {
	ev_real = real ( eval [ i ] ) ;
	// is the eigval less than zero?
	if ( abs ( ev_real ) < pow ( 10 , -6 ) ) {
		count = 0 ;
		  // count the number of elements in eigvec who's eigval = 0; 
    		  for ( int j = 0 ; j < order ; j++ ) {
        	    if ( abs ( real ( evec[ j * n + i ] ) ) < pow ( 10 , -6 ) ) { count++ ;}  
    		  }
		// if eigvec elements have more than 1 real component use it
          	if ( count <=  1 ) { k = i ; }  
	  }
      }

  for ( int j = 0 ; j < order ; j++ ) {
	sum = sum + evec[ j * n + k ] ;
  }

  for ( int j = 0 ; j < order ; j++ ) {
	M0[ j ] = evec[ j * n + k ] / sum ;
  }
}




#endif
